/*
  wolfinfo.c  --  retrieve information from rtcw-server/master
  Copyright (C) 2002 Steffen Solyga <solyga@absinth.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include	"wolfinfo.h"


int eno;			/* error number for some functions */
int verbose= 0;
char* pn= NULL;			/* prg_name is Makefile constant */



int
display_help( char* pn ) {
  fprintf( HELP_CHANNEL, "%s v%s (%s): ", pn, VERSION_NUMBER, DATE_OF_LAST_MOD );
  fprintf( HELP_CHANNEL, "Get infos from RtCW-Server/Master.\n" );
  fprintf( HELP_CHANNEL, "Flowers & bug reports to %s.\n", MY_EMAIL_ADDRESS );
  fprintf( HELP_CHANNEL, "Usage: %s [options] [host[:port]]\n", pn );
  fprintf( HELP_CHANNEL, "switches:\n");
  fprintf( HELP_CHANNEL, "  -h\t help, write this info to %s and exit sucessfully\n", HELP_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "  -l\t list hosts (don't contact them)\n" );
  fprintf( HELP_CHANNEL, "  -L\t list hosts which could be contacted\n" );
  fprintf( HELP_CHANNEL, "  -v\t increase verbosity level (on %s) by one\n", VERBOSE_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "  -V\t print version and compilation info to %s and exit sucessfully\n", VERSION_CHANNEL==stdout?"stdout":"stderr" );
  fprintf( HELP_CHANNEL, "controllers:\n");
  fprintf( HELP_CHANNEL, "  -a attempts\t set minimum number of attempts (default %d)\n", DEFAULT_ATTEMPTS );
  fprintf( HELP_CHANNEL, "  -c mode\t set color mode (default %d)\n", DEFAULT_COLOR_MODE );
  fprintf( HELP_CHANNEL, "  -f format\t set print format (default %s)\n", DEFAULT_PFORMAT );
  fprintf( HELP_CHANNEL, "  -m master\t set master server (default %s:%d)\n", DEFAULT_MNAME, DEFAULT_MPORT );
  fprintf( HELP_CHANNEL, "  -o protocol\t set protocol version (default %d)\n", DEFAULT_PROTO );
  fprintf( HELP_CHANNEL, "  -p port\t set local port (default %d)\n", DEFAULT_LPORT );
  fprintf( HELP_CHANNEL, "  -t time\t set timeout (in seconds) for game server responses (default %d)\n", DEFAULT_STIMEOUT );
  return( 0 );
}


int
display_version( char* pn ) {
  fprintf( VERSION_CHANNEL, "%s v%s (%s)\n", pn, VERSION_NUMBER, DATE_OF_LAST_MOD );
  fprintf( VERSION_CHANNEL, "compilation settings:\n" );
  fprintf( VERSION_CHANNEL, "  DEFAULT_MNAME     : %s\n", DEFAULT_MNAME );
  fprintf( VERSION_CHANNEL, "  DEFAULT_MPORT     : %d\n", DEFAULT_MPORT );
  fprintf( VERSION_CHANNEL, "  DEFAULT_SNAME     : %s\n", DEFAULT_SNAME );
  fprintf( VERSION_CHANNEL, "  DEFAULT_SPORT     : %d\n", DEFAULT_SPORT );
  fprintf( VERSION_CHANNEL, "  DEFAULT_LPORT     : %d\n", DEFAULT_LPORT );
  fprintf( VERSION_CHANNEL, "  DEFAULT_PROTO     : %d\n", DEFAULT_PROTO );
  fprintf( VERSION_CHANNEL, "  DEFAULT_MTIMEOUT  : %d seconds\n", DEFAULT_MTIMEOUT );
  fprintf( VERSION_CHANNEL, "  DEFAULT_STIMEOUT  : %d seconds\n", DEFAULT_STIMEOUT );
  fprintf( VERSION_CHANNEL, "  DEFAULT_ATTEMPTS  : %d\n", DEFAULT_ATTEMPTS );
  fprintf( VERSION_CHANNEL, "  DEFAULT_COLOR_MODE: %d\n", DEFAULT_COLOR_MODE );
  fprintf( VERSION_CHANNEL, "  BUF_SIZE          : %d\n", BUF_SIZE );
  fprintf( VERSION_CHANNEL, "  DEFAULT_PFORMAT   : %s\n", DEFAULT_PFORMAT );
  return( 0 );
}



char*
str_dsc( const char* str1, const char* str2) {
/*
returns pointer to first char of str1 not contained in str2
started 1997-02-09
*/
  char *p1, *p2;
  for ( p1=(char*)str1; ; p1++ ) {
    if ( *p1 == '\0' ) return( p1 );
    for ( p2=(char*)str2; ; p2++ ) {
      if ( *p2 == '\0' ) return( p1 );
      if ( *p2 == *p1 ) break;
    }
  }
}


int
str_stp( char* str1, char* str2 ) {
/*
writes '\0' to first char in str1 matching one of the chars from str2,
returns number of deleted characters
started 1997-03-28
*/
  char *p1, *p2;
  int n= 0;
  for( p1=str1; ; p1++ ) {
    if ( *p1 == '\0' ) return( n );
    for( p2=str2; ; p2++ ) {
      if( *p2 == '\0' ) break;
      if( *p2 == *p1 ) {
        *p1= '\0';
        while( *(p1+++n) != '\0' );
        return( n );
      }
    }
  }
}



void*
memmem( const void* haystack, size_t haystacklen, \
        const void *needle, size_t needlelen ) {
/*
 * find needle in haystack 
 * return start address of needle or
 *        NULL on "no such needle in haystack"
 * Hint: memmem() is a GNU extension to libc
 * started: Thu Feb 14 01:39:55 CET 2002 @beast
 */
  void* p;
  size_t l;
  for( p=(void*)haystack, l=haystacklen; l>=needlelen; p++, l-- )
    if( memcmp(p,needle,needlelen) == 0 ) return( p );
  return( NULL );
}



char*
basename( char* name ) {
/*
 * strip directory from name
 * returns pointer to stripped name
 * hint: basename("/usr/bin/") == ""
 *   basename(1) would return "bin" !!
 */
  char* p= name;
  while( *p != '\0' ) p++;
  while( p > name ) {
    if( *(p-1) == '/' ) break;
    else p--;
  }
  return( p );
}



in_addr_t
get_addr_by_name( char* hostname ) {
/* return IPv4 address in network byte order */
  struct hostent* ent= gethostbyname( hostname );
  if( ent == NULL ) return( -1 );
  if( ent->h_addr_list[0] == NULL ) return( -1 );
  if( ent->h_addrtype != AF_INET ) {
    h_errno= 5;                         /* no IPv4 address, my extension */
    return( -1 );
  }
  return( *((in_addr_t*)ent->h_addr_list[0]) );
}



void
free_host( struct host* a_host ) {
/* free all host memory */
  if( a_host != NULL ) {
    free( a_host->name );
    free( a_host->info );
    free( a_host->stat );
    free( a_host );
  }
  return;
}



void
free_hlist( struct host* hlist ) {
/* free all list memory */
  struct host* n_host;
  struct host* a_host;
  for( a_host=hlist; a_host!=NULL; ) {
    n_host= a_host->next;
    free_host( a_host );
    a_host= n_host;
  }
  return;
}



struct host*
create_host( void ) {
/* create empty host structure */
  struct host* a_host= NULL;
  if( (a_host=malloc(sizeof(struct host))) == NULL ) return( NULL );
  a_host->next= NULL; 
  a_host->where= NULL;
  a_host->name= NULL;
  a_host->sock.sin_family= AF_INET;
  a_host->sock.sin_port= (in_port_t) 0;
  a_host->sock.sin_addr.s_addr= (in_addr_t) 0;
  a_host->info= NULL;
  a_host->stat= NULL;
  return( a_host );
}


struct host*
create_host_from_name( const char* name, char** endptr, \
                       const char* default_name, int default_port ) {
/*
 * create host structure for single host from name
 * set members name and sock.sin_port (use defaults if necessary)
 * return pointer to new host structure or NULL on error (eno)
 * *endptr points to first unprocessed char in name
 */
  struct host* a_host= NULL;
  char* start= NULL;
  char* end= NULL;
  eno= ENO_OK;
  *endptr= (char*) name;
  if( name == NULL ) {
    fprintf( ERROR_CHANNEL, "%s: create_host_from_name(): INVCALL bug!\n", pn );
    eno= ENO_INVCALL; goto RETURN_CREATE_HOST_FROM_NAME;
  }
  if( (a_host=create_host()) == NULL ) {
    eno= ENO_MALLOC; goto RETURN_CREATE_HOST_FROM_NAME;
  }
  /* set hostname */
  {
    char* s;
    size_t len;
    char* p;
    start= str_dsc( name, " \t,;" );
    end= strpbrk( start, ": \t,;" );
    if( end == NULL ) end= start + strlen( start );
    if( endptr != NULL ) *endptr= end;
    s= start;
    len= end - start;
    if( len == 0 ) {
      if( default_name == NULL ) {
        eno= ENO_INVCALL; goto RETURN_CREATE_HOST_FROM_NAME;
      }
      s= (char*) default_name;
      len= strlen( default_name );
    }
    if( (p=malloc(len+1)) == NULL ) {
      eno= ENO_MALLOC; goto RETURN_CREATE_HOST_FROM_NAME;
    }
    strncpy( p, s, len );
    p[len]= '\0';
    a_host->name= p;
  }
  /* set port */
  a_host->sock.sin_port= htons(default_port);
  if( *end == ':' ) {
    long port_l;
    start= ++end;
    port_l= strtol( start, &end, 0 );
    if( end != start ) {
      if( port_l<1 || port_l>65535 ) {
        fprintf( ERROR_CHANNEL, "%s: Invalid port %ld.\n", pn, port_l );
        eno= ENO_INVPORT; goto RETURN_CREATE_HOST_FROM_NAME;
      }
      a_host->sock.sin_port= htons((in_port_t)port_l);
    }
    if( endptr != NULL ) *endptr= end;
  }
RETURN_CREATE_HOST_FROM_NAME:
  if( eno != ENO_OK ) {
    free_hlist( a_host );
    a_host = NULL;
  }
  return( a_host );
}



struct host*
create_hlist_from_name( const char* name, \
                        const char* default_name, \
                        in_port_t default_port ) {
/*
 * create linked host structures from name string
 * set members name and sock.sin_port (use defaults if necessary)
 * return pointer to first host in list or NULL on error (eno)
 */
  struct host* hlist= NULL;
  struct host* new_host= NULL;
  struct host* old_host= NULL;
  char* s= (char*) name;
  eno= ENO_OK;
  if( name == NULL ) goto RETURN_CREATE_HLIST_FROM_NAME;
  while( *(s=str_dsc(s," \t,;")) != '\0' ) {
    new_host= create_host_from_name( s, &s, default_name, default_port );
    if( new_host == NULL ) goto RETURN_CREATE_HLIST_FROM_NAME;
    if( hlist == NULL ) hlist= new_host;
    if( old_host != NULL ) old_host->next= new_host;
    old_host= new_host;
  }
RETURN_CREATE_HLIST_FROM_NAME:
  if( eno != ENO_OK ) {
    free_hlist( hlist );
    hlist = NULL;
  }
  return( hlist );
}



struct host*
extend_hlist( struct host* hlist, struct host* new_list ) {
/* append new_list to hlist */
  struct host* i= hlist;
  if( hlist == NULL ) return( new_list );
  while( i->next != NULL ) i= i->next; 
  i->next= new_list;
  return( hlist );
}



struct host*
remove_host_from_hlist( struct host* hlist, struct host* a_host ) {
/*
 * ... and free all host memory
 */
  if( hlist == NULL ) return( NULL );
  if( a_host != NULL ) {
    struct host* n= a_host->next;	/* next of a_host */
    struct host* i;
    struct host* p;			/* prev of i */
    free_host( a_host );
    if( a_host == hlist ) return( n );
    for( i=hlist, p=hlist; i!=NULL; p=i, i=i->next ) {
      if( i == a_host ) {
        p->next= n;
        break;
      }
    }
    /* a_host not in hlist */
  }
  return( hlist );
}



struct host*
correct_hlist( struct host* hlist ) {
/*
 * remove repeatedly occuring hosts from hlist
 * names take precedence
 */
  struct host* i;
  struct host* j;
STARTOVER_CORRECT_HLIST:
  for( j=hlist; j!=NULL; j=j->next ) {
    for( i=j->next; i!=NULL; i=i->next ) {
      if( memcmp(&i->sock,&j->sock,sizeof(struct sockaddr_in)) == 0 ) {
        /* equal peers or two unresolved hosts with equal ports */
        if( i->name == NULL ) {
          hlist= remove_host_from_hlist( hlist, i );
          goto STARTOVER_CORRECT_HLIST;
        }
        else if( j->name == NULL ) {
          hlist= remove_host_from_hlist( hlist, j );
          goto STARTOVER_CORRECT_HLIST;
        }
        else if( strcmp(i->name,j->name) == 0 ) {
          hlist= remove_host_from_hlist( hlist, j );
          goto STARTOVER_CORRECT_HLIST;
        }
        else { /* hmm, different names but same peer... */
          if( *str_dsc(i->name," .0123456789") == '\0' ) {
            hlist= remove_host_from_hlist( hlist, i );
            goto STARTOVER_CORRECT_HLIST;
          }
          if( *str_dsc(j->name," .0123456789") == '\0' ) {
            hlist= remove_host_from_hlist( hlist, j );
            goto STARTOVER_CORRECT_HLIST;
          }
        }
      }
    }
  }
  return( hlist );
}



struct host*
resolve_hlist( struct host* hlist ) {
  struct host* a_host= hlist;
  struct host* next;
  eno= ENO_OK;
  for( a_host=hlist; a_host!=NULL; a_host=next ) {
    next= a_host->next;
    if( a_host->name != NULL  &&  a_host->sock.sin_addr.s_addr == 0 ) {
      /* gethostbyname */
      in_addr_t addr;
      if( (addr=get_addr_by_name( a_host->name )) == -1 ) {
        fprintf( WARN_CHANNEL, "%s: Ignoring '%s'. %s.\n",
                 pn, a_host->name, h_errlist[h_errno] );
        hlist= remove_host_from_hlist( hlist, a_host );
        continue;
      }
      a_host->sock.sin_addr.s_addr= addr;
    }
    else if(a_host->name == NULL  &&  a_host->sock.sin_addr.s_addr != 0 ) {
      /* generate name */
      char* p;
      if( (p=malloc(INET_ADDRSTRLEN)) == NULL ) {
        fprintf( VERBOSE_CHANNEL, "%s: Malloc() error, cleaning up.\n", pn );
        free_hlist( hlist );
        eno= ENO_MALLOC; return( NULL );
      }
      inet_ntop( AF_INET, &a_host->sock.sin_addr, p, INET_ADDRSTRLEN );
      a_host->name= p;
    }
    else if( a_host->name == NULL  &&  a_host->sock.sin_addr.s_addr == 0 ) {
      /* empty host structure */
      hlist= remove_host_from_hlist( hlist, a_host );
    }
    /* host already resolved */
  }
  return( hlist );
}



int
count_hosts_from_master( struct host* hlist, struct host* master ) {
  struct host* a_host= hlist;
  int noh= 0;
  for( a_host=hlist; a_host!=NULL; a_host=a_host->next )
    if( a_host->where == master ) noh++;
  return( noh );
}



int 
count_hosts( struct host* hlist ) {
  struct host* a_host= hlist;
  int noh= 0;
  for( a_host=hlist; a_host!=NULL; a_host=a_host->next ) noh++;
  return( noh );
}



void
fprint_hlist( FILE* channel, const char* head, struct host* hlist ) {
  struct host* a_host= hlist;
  char default_head[]= "";
  char default_name[]= "(null)";
  FILE* c= channel;
  char* h= (char*) head;
  char* n;
  if( channel == NULL ) c= VERBOSE_CHANNEL;
  if( head == NULL ) h= default_head;
  for( a_host=hlist; a_host!=NULL; a_host=a_host->next ) {
    if( (n=a_host->name) == NULL ) n= default_name;
    fprintf( c, "%s%s:%d\n", h, n, ntohs(a_host->sock.sin_port) );
  }
  return;
}



int
create_socket( int port, struct timeval* timeout, socklen_t len ) {
  int fd;
  struct sockaddr_in addr;
  addr.sin_addr.s_addr= htonl(INADDR_ANY);
  addr.sin_port= htons(port);
  if( (fd=socket(AF_INET,SOCK_DGRAM,0)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Cannot create socket. %s.\n",
             pn, sys_errlist[errno] );
    return( -1 );
  }
  if( bind(fd,(struct sockaddr*)&addr,sizeof(addr)) == -1 ) {
    fprintf( WARN_CHANNEL, "%s: Cannot bind port %d to socket. %s.\n",
             pn, port, sys_errlist[errno] );
  }
  if( setsockopt(fd,SOL_SOCKET,SO_RCVTIMEO,timeout,sizeof(struct timeval)) == -1 ) {
    fprintf( ERROR_CHANNEL, "%s: Cannot set timeout for socket. %s.\n",
             pn, sys_errlist[errno] );
    return( -1 );
  }
  return( fd );
}



struct host*
create_hlist_from_mrep( const char* buf, int size, \
                        struct host* master  ) {
/*
 * create hostlist from master reply
 * return hostlist or NULL on error (eno set)
 * (valid frame with 0 addresses is an error - ENO_MALEFMREP)
 *
 * each datagram expected has at most 810 (0x032a) bytes,
 * starts with "XXXXgetserversResponse" (X=0xff)
 * continues with sets of '\',IP,PORT (1+4+2 bytes)
 * and ends with "\EOT"
 * ==> at most 112 server addresses
 */
  char ans[]= "XXXXgetserversResponse";
  struct host* hlist= NULL;
  struct host* old_host= NULL;
  int tara= strlen(ans) + 1 + strlen(END_STR);
  int netto= size - tara;
  int noa= netto/7;				/* number of addresses */
  char* bp;
  char* bp_max;
  *(uint32_t*)ans= MARKER;
  eno= ENO_OK;
  /* check vars */
  if( buf == NULL  ||  size < 0 ) {
    eno= ENO_INVCALL; goto RETURN_CREATE_HLIST_FROM_MREP;
  }
  /* check buffer */
  if( netto < 0 ) {
    eno= ENO_MALEFMREP; goto RETURN_CREATE_HLIST_FROM_MREP;
  }
  if( memcmp(buf,ans,strlen(ans)) != 0 ) {
    eno= ENO_MALEFMREP; goto RETURN_CREATE_HLIST_FROM_MREP;
  }
  if( buf[size-(1+strlen(END_STR))] != ESC_CHAR ) {
    eno= ENO_MALEFMREP; goto RETURN_CREATE_HLIST_FROM_MREP;
  }
  if( memcmp(buf+size-strlen(END_STR),END_STR,strlen(END_STR)) != 0 ) {
    eno= ENO_MALEFMREP; goto RETURN_CREATE_HLIST_FROM_MREP;
  }
  if( netto%7 ) {
    eno= ENO_MALEFMREP; goto RETURN_CREATE_HLIST_FROM_MREP;
  }
  /* get addresses */
  for( bp=(char*)buf+strlen(ans), bp_max=bp+7*noa; bp<bp_max; bp+=7 ) {
    struct host* new_host;
    if( *bp != ESC_CHAR ) {
      eno= ENO_MALEFMREP; goto RETURN_CREATE_HLIST_FROM_MREP;
    }
    if( (new_host=create_host()) == NULL ) {
      eno= ENO_MALLOC; goto RETURN_CREATE_HLIST_FROM_MREP;
    }
    new_host->where= master;
    new_host->sock.sin_addr.s_addr= *(in_addr_t*)(bp+1);
    new_host->sock.sin_port= *(in_port_t*)(bp+5);
    if( hlist == NULL ) hlist= new_host;
    if( old_host != NULL ) old_host->next= new_host;
    old_host= new_host;
  }
RETURN_CREATE_HLIST_FROM_MREP:
  if( eno != ENO_OK ) {
    free_hlist( hlist );
    hlist= NULL;
  }
  return( hlist );
}



int
get_int_from_srep( const char* buf, int size, const char* valname ) {
  long val;
  char* p;
  char* pold;
  if( (p=memmem(buf,size,valname,strlen(valname))) == NULL ) return( -1 );
  p+= strlen(valname);
  pold= p;
  val= strtol( p, &p, 10 );
  if( p == pold  ||  val < 0  ||  val > INT_MAX ) return( -1 );
  return( (int) val );
}



char*
get_str_from_srep( const char* buf, int size, const char* valname ) {
  char* p;
  char* s;
  size_t len;
  if( (p=memmem(buf,size,valname,strlen(valname))) == NULL ) return( NULL );
  p+= strlen(valname);
  if( (s=memchr(p,'\\',size-(p-buf))) != NULL ) len= s-p;
  else if( (s=memchr(p,'\n',size-(p-buf))) != NULL ) len= s-p;
  else len= size-(p-buf);
  if( (s=malloc(len+1)) == NULL ) return( NULL );
  memcpy(s,p,len);
  s[len]= '\0';
  return( s );
}



void
fprint_colored_str( FILE* channel, int color_mode, \
                const char* start, const char* end ) {
  char* p;
  if( end <= start ) return;
  for( p=(char*)start; p<end; p++ ) {
    unsigned char c= *p;
    if( color_mode>0 && c==COLOR_ESC_CHAR ) {
      c= *++p;
      if( c == '\0' ) {
        fprintf( channel, "%c", COLOR_ESC_CHAR );
      }
      else if( c == COLOR_ESC_CHAR ) {
        fprintf( channel, "%c", COLOR_ESC_CHAR );
        p--;
      }
      else if( color_mode >= 2 ) {
        c= (c&0x07);
        if( color_mode > 2 ) c+= 8;
        fprintf( channel, "\e[%sm", color_table[c] );
      }
    }
    else if( c<0x20 || (c>0x7e && c<0xa0) ) {
      fprintf( channel, "(0x%02x)", c );
    }
    else {
      fprintf( channel, "%c", c );
    }
  }
  if( color_mode >= 2 ) fprintf( channel, "\e[00m" );
  return;
}



void
fprint_str_from_resp( FILE* channel, int color_mode, \
                      const char* resp, const char* varname ) {
  FILE* ch= channel;
  char* s;
  char* p;
  if( ch == NULL ) ch= STD_CHANNEL;
  if( resp==NULL || varname==NULL ) {
    fprintf( ch, "(unknown)" );
    return;
  }
  if( (s=strstr(resp,varname)) == NULL ) {
    fprintf( ch, "(undefined)" );
    return;
  }
  s+= strlen( varname );
  for( p=s; *p!='\0' && *p!='\\' && *p!='\n'; p++ );
  fprint_colored_str( ch, color_mode, s, p );
  return;
}



void
fprint_players_from_resp( FILE* channel, int color_mode, const char* resp ) {
  FILE* ch= channel;
  char* s;
  char* p;
  int n;
  if( ch == NULL ) ch= STD_CHANNEL;
  if( resp == NULL ) {
    fprintf( ch, "(unknown)" );
    return;
  }
  /* find first player */
  for( s=(char*)resp; *s!='\0'; s++ ) if( *s == '\n' ) break;
  if( *s == '\0' ) return;
  s++;
  /* print players */
  for( p=s,n=0; *p!='\0'; p++ ) {
    if( *p == '"' ) { 
      n++;
      if( n%2 ) {
        s= p;
      }
      else {
        fprint_colored_str( ch, color_mode, s, p+1 );
        fprintf( ch, " " );
      }
    }
  }
  return;
}



void
fprintf_response( FILE* channel, const char* format, \
                  int c_mode, struct host* server ) {
  FILE* ch= channel;
  char* fp;			/* format pointer */
  char* rp;			/* response pointer */
  int bs;
  int c;
  int i;
  if( ch==NULL ) ch= STD_CHANNEL;
  if( format==NULL || server==NULL ) return;
  if( server->info==NULL && server->stat==NULL ) return;
  /* make variable names lower case */
  if( server->info != NULL ) {
    for( rp=server->info,bs=0; *rp!='\0'; rp++ ) {
      if( *rp == '\n' ) break;
      if( *rp == '\\' ) bs++;
      if( bs%2  &&  *rp != '\\' ) *rp= tolower(*rp);
    }
  }
  if( server->stat != NULL ) {
    for( rp=server->stat,bs=0; *rp!='\0'; rp++ ) {
      if( *rp == '\n' ) break;
      if( *rp == '\\' ) bs++;
      if( bs%2  &&  *rp != '\\' ) *rp= tolower(*rp);
    }
  }
  /* process format string and buffer */
  for( fp=(char*)format; (c=*fp)!='\0'; fp++ ) {
    if( c == '%' ) { /* special wolf format sequence */
      fp++;
      switch( *fp ) {
        case '\0': fp--;
        case '%' : fprintf( ch, "%%" );
          break;
        /* general cases */
        case GFORMAT_HOST:
          fprintf( ch, "%s", server->name );
          break;
        case GFORMAT_PORT:
          fprintf( ch, "%d", htons(server->sock.sin_port) );
          break;
        case GFORMAT_LINE:
          for( i=0; i<80; i++ ) fprintf( ch, "-" );
          fprintf( ch, "\n" );
          break;
        /* info cases */
        /* remember to add info cases in format_requires_info() also !! */
        case IFORMAT_ALL:
          fprintf( ch, "%s", server->info );
          break;
        case IFORMAT_PROTO:
          fprint_str_from_resp( ch, 0, server->info, IRESP_PROTO );
          break;
        case IFORMAT_SERVER:
          fprint_str_from_resp( ch, c_mode, server->info, IRESP_SERVER );
          break;
        case IFORMAT_MAP:
          fprint_str_from_resp( ch, 0, server->info, IRESP_MAP );
          break;
        case IFORMAT_CLIENTS:
          fprint_str_from_resp( ch, 0, server->info, IRESP_CLIENTS );
          break;
        case IFORMAT_MAXCLIENTS:
          fprint_str_from_resp( ch, 0, server->info, IRESP_MAXCLIENTS );
          break;
        case IFORMAT_GAMETYPE:
          fprint_str_from_resp( ch, 0, server->info, IRESP_GAMETYPE );
          break;
        case IFORMAT_GAMESKILL:
          fprint_str_from_resp( ch, 0, server->info, IRESP_GAMESKILL );
          break;
        case IFORMAT_FRIENDLY:
          fprint_str_from_resp( ch, 0, server->info, IRESP_FRIENDLY );
          break;
        case IFORMAT_MAXLIVES:
          fprint_str_from_resp( ch, 0, server->info, IRESP_MAXLIVES );
          break;
        case IFORMAT_TOURNEY:
          fprint_str_from_resp( ch, 0, server->info, IRESP_TOURNEY );
          break;
        case IFORMAT_PUNK:
          fprint_str_from_resp( ch, 0, server->info, IRESP_PUNK );
          break;
        /* status cases */
        /* remember to add status cases in format_requires_status() also !! */
        case SFORMAT_ALL:
          fprintf( ch, "%s", server->stat );
          break;
        case SFORMAT_VERSION:
          fprint_str_from_resp( ch, 0, server->stat, SRESP_VERSION );
          break;
        case SFORMAT_TIMELIMIT:
          fprint_str_from_resp( ch, 0, server->stat, SRESP_TIMELIMIT );
          break;
        case SFORMAT_PLAYERS:
          fprint_players_from_resp( ch, c_mode, server->stat );
          break;
        default:
          fprintf( ch, "%%" );
          fp--;
      }
    }
    else if( c == '\\' ) { /* usual ctrl sequence */
      fp++;
      switch( *fp ) {
        case '\0': fp--;
        case '\\': fprintf( ch, "\\" ); break;
        case 'a' : fprintf( ch, "\a" ); break;
        case 'b' : fprintf( ch, "\b" ); break;
        case 't' : fprintf( ch, "\t" ); break;
        case 'n' : fprintf( ch, "\n" ); break;
        case 'v' : fprintf( ch, "\b" ); break;
        case 'f' : fprintf( ch, "\f" ); break;
        case 'r' : fprintf( ch, "\r" ); break;
        default  : fprintf( ch, "\\%c", *fp );
      }
    }
    else {
      fprintf( ch, "%c", *fp );
    }
  }
  return;
}



int
format_requires_info( const char* format ) {
  char* fp;			/* format pointer */
  int c;
  for( fp=(char*)format; (c=*fp)!='\0'; fp++ ) {
    if( c == '%' ) { /* special wolf format sequence */
      fp++;
      switch( *fp ) {
        case '\0': fp--;
        case '%' : break;
        /* ALL info cases */
        /* remember to add info cases in fprintf_response() also !! */
        case IFORMAT_ALL:
        case IFORMAT_PROTO:
        case IFORMAT_SERVER:
        case IFORMAT_MAP:
        case IFORMAT_CLIENTS:
        case IFORMAT_MAXCLIENTS:
        case IFORMAT_GAMETYPE:
        case IFORMAT_GAMESKILL:
        case IFORMAT_FRIENDLY:
        case IFORMAT_MAXLIVES:
        case IFORMAT_TOURNEY:
        case IFORMAT_PUNK:
          return( 1 ); break;
        default: fp--;
      }
    }
  }
  return( 0 );
}



int
format_requires_status( const char* format ) {
  char* fp;			/* format pointer */
  int c;
  for( fp=(char*)format; (c=*fp)!='\0'; fp++ ) {
    if( c == '%' ) { /* special wolf format sequence */
      fp++;
      switch( *fp ) {
        case '\0': fp--;
        case '%' : break;
        /* ALL status cases */
        /* remember to add status cases in fprintf_response() also !! */
        case SFORMAT_ALL:
        case SFORMAT_VERSION:
        case SFORMAT_TIMELIMIT:
        case SFORMAT_PLAYERS:
          return( 1 ); break;
        default: fp--;
      }
    }
  }
  return( 0 );
}



int
main( int argc, char** argv ) {
/*
 * main() wolfinfo
 * started      : Sun Jan 27 19:14:30 CET 2002 @beast
 */
  char* fpn= *argv;
  int c;
  int retval= RETVAL_OK;
  struct host* mlist= NULL;		/* master server list */
  struct host* slist= NULL;		/* game server list */
  int sock_fd;
  int list_hosts= 0;			/* 3-state flag list hosts */
  int contact_masters= 1;		/* switch contact master servers */
  int get_info= 1;			/* switch getinfo from game servers */
  int get_status;			/* switch getstatus from game servers */
  int attempts= DEFAULT_ATTEMPTS;	/* attempts to get info */
  int color_mode= DEFAULT_COLOR_MODE;
  char default_pformat[]= DEFAULT_PFORMAT;
  char* pformat= default_pformat;
  struct timeval mtimeout= { (long)DEFAULT_MTIMEOUT, 0L };
  struct timeval stimeout= { (long)DEFAULT_STIMEOUT, 0L };
  int lport= DEFAULT_LPORT;		/* local port */
  int proto= DEFAULT_PROTO;
  char buf[BUF_SIZE];

/* process options */
  pn= basename( fpn );
  *argv= pn;				/* give getop() the cut name */
  while( (c=getopt(argc,argv,"a:c:f:hlLm:o:p:t:vV")) != EOF ) {
    switch( c ) {
      case 'a': /* set number of attempts to get info from game server */ {
        long val;
        char* p= optarg;
        val= strtol( str_dsc(p," \t"), &p, 0 );
        if( p == optarg || val < 1 ) {
          fprintf( ERROR_CHANNEL, "%s: Invalid number of attempts '%s'.\n",
                   pn, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        attempts= (int) val;
        break;
      }
      case 'c': /* set color mode */ {
        long val;
        char* p= optarg;
        val= strtol( str_dsc(p," \t"), &p, 0 );
        if( p == optarg || val < 0 || val > 3 ) {
          fprintf( ERROR_CHANNEL, "%s: Invalid color mode '%s'.\n",
                   pn, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        color_mode= (int) val;
        break;
      }
      case 'f': /* set print format */
        pformat= optarg;
        break;
      case 'h': /* display help and exit sucessfully */
        display_help( pn );
        retval= RETVAL_OK; goto DIE_NOW;
      case 'l': /* output host list (from master(s) and specified) */
        list_hosts= 1;
        break;
      case 'L': /* output host list (from master(s) and specified) */
        list_hosts= 2;
        break;
      case 'm': /* set master server(s) */ {
        struct host* alist;
        alist= create_hlist_from_name( optarg, DEFAULT_MNAME, DEFAULT_MPORT );
        if( alist == NULL ) {
          retval= eno==ENO_INVCALL ? RETVAL_BUG : RETVAL_ERROR;
          goto DIE_NOW;
        }
        mlist= extend_hlist( mlist, alist );
        break;
      }
      case 'o':	/* protocol number */ {
        long val;
        char* p= optarg;
        val= strtol( str_dsc(p," \t"), &p, 0 );
        if( p == optarg || val < 1 || val > 99 ) {
          fprintf( ERROR_CHANNEL, "%s: Invalid protocol number '%s'.\n",
                   pn, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        proto= (int) val;
        break;
      }
      case 'p': /* local port */ {
        long val;
        char* p= optarg;
        val= strtol( str_dsc(p," \t"), &p, 0 );
        if( p == optarg || val < 1 || val > 65535 ) {
          fprintf( ERROR_CHANNEL, "%s: Invalid port '%s'.\n", pn, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        lport= (uint16_t) val;
        break;
      }
      case 't': /* game server timeout in seconds */ {
        long val;
        char* p= optarg;
        val= strtol( str_dsc(p," \t"), &p, 0 );
        if( p == optarg || val < 0 ) {
          fprintf( ERROR_CHANNEL, "%s: Timeout '%s' invalid.\n", pn, optarg );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        stimeout.tv_sec= val;
        break;
      }
      case 'v': /* increase verbosity level */
        verbose++;
        break;
      case 'V': /* display version to VERSION_CHANNEL and exit sucessfully */
        display_version( pn );
        retval= RETVAL_OK; goto DIE_NOW;
      case '?': /* refer to -h and exit unsucessfully */
        fprintf( ERROR_CHANNEL, "%s: Try '%s -h' for more information.\n",
                 pn, pn );
        retval= RETVAL_ERROR; goto DIE_NOW;
      default : /* program error */
        fprintf( ERROR_CHANNEL, "%s: Options bug! E-mail me at %s.\n",
                 pn, MY_EMAIL_ADDRESS );
        retval= RETVAL_BUG; goto DIE_NOW;
    }
  }

/* set up game server list from cmd-line args */
  for( ; optind<argc; optind++ ) {
    struct host* alist;
    alist= create_hlist_from_name( argv[optind], DEFAULT_SNAME, DEFAULT_SPORT );
    if( alist == NULL ) {
      retval= eno==ENO_INVCALL ? RETVAL_BUG : RETVAL_ERROR;
      goto DIE_NOW;
    }
    slist= extend_hlist( slist, alist );
    if( !list_hosts ) contact_masters= 0;
  }

/* resolve game server list */
  if( verbose >= 2 ) {
    fprintf( VERBOSE_CHANNEL, "%s: Resolving game servers.\n", pn );
  }
  slist= resolve_hlist( slist );

/* retrieve/extend game server list from master(s) */
  if( contact_masters ) {
    struct host* master;
    int tnbs, tnbr;
    int nsg= 112;		/* number of servers got */
    if( mlist == NULL ) { 
      mlist= create_hlist_from_name( ":", DEFAULT_MNAME, DEFAULT_MPORT );
      if( mlist == NULL ) {
        retval= eno==ENO_INVCALL ? RETVAL_BUG : RETVAL_ERROR;
        goto DIE_NOW;
      }
    }
    if( (mlist=resolve_hlist(mlist)) == NULL ) {
      fprintf( ERROR_CHANNEL, "%s: No valid master server known.\n", pn );
      retval= RETVAL_ERROR; goto DIE_NOW;
    }
    if( (sock_fd=create_socket(lport,&mtimeout,sizeof(mtimeout))) == -1 ) {
      retval= RETVAL_ERROR; goto DIE_NOW;
    }
    if( verbose >= 1  ) {
      fprintf( VERBOSE_CHANNEL, "%s: %s game server list.\n",
               pn, slist==NULL?"Retrieving":"Extending" );
    }
    tnbs= 0; tnbr= 0;
    /* contact master(s) */
    for( master=mlist; master!=NULL; master=master->next ) {
      int nbs;
      char req[]= "XXXXgetservers 00 full empty";
      snprintf( req, strlen(req)+1, "%s %02d %s %s",
               "XXXXgetservers", proto, "full", "empty" );
      *(uint32_t*)req= MARKER;
      if( verbose >= 2 ) {
        fprintf( VERBOSE_CHANNEL, "%s: Contacting %s:%d.\n",
                 pn, master->name, ntohs(master->sock.sin_port) );
      }
      nbs= sendto(sock_fd,req,strlen(req),0, \
                   (struct sockaddr*)&master->sock,sizeof(master->sock) );
      if( nbs == -1 ) {
        fprintf( ERROR_CHANNEL, "%s: Sendto() failed for %s:%d. %s.\n",
                 pn, master->name, ntohs(master->sock.sin_port),
                 sys_errlist[errno] );
      }
      else {
        tnbs+= nbs;
      }
    }
    /* listen for replies, set up game server list */
    while( nsg >= 112 ) {
      struct sockaddr_in ans_addr;
      socklen_t socklen= sizeof( ans_addr );
      int nbr;
      struct host* alist;
      nbr= recvfrom( sock_fd, buf, BUF_SIZE, 0, \
                     (struct sockaddr*)&ans_addr, &socklen );
      if( nbr == -1 ) {
        if( errno==EAGAIN ) {
          if( tnbr == 0 )
            fprintf( WARN_CHANNEL, "%s: No response from master(s).\n", pn );
          break;
        }
        fprintf( ERROR_CHANNEL, "%s: Recvfrom() failed. %s.\n",
                 pn, sys_errlist[errno] );
        retval= RETVAL_ERROR; goto DIE_NOW;
      }
      for( master=mlist; master!=NULL; master=master->next ) {
        if( memcmp(&master->sock,&ans_addr,socklen) == 0 ) break;
      }
      if( master == NULL ) {
        char ans_name[INET_ADDRSTRLEN]= "";
        int ans_port= ntohs( ans_addr.sin_port );
        inet_ntop( AF_INET, &ans_addr.sin_addr, ans_name, INET_ADDRSTRLEN );
        fprintf( WARN_CHANNEL, "%s: Ignoring dgram from %s:%d. %s.\n", pn,
                 ans_name, ans_port, "Master unknown"  );
        continue;
      }
      if( nbr == BUF_SIZE ) {
        fprintf( ERROR_CHANNEL, "%s: Buffer too small. E-mail me at %s.\n",
                 pn, MY_EMAIL_ADDRESS );
        retval= RETVAL_BUG; goto DIE_NOW;
      }
      tnbr+= nbr;
      if( (alist=create_hlist_from_mrep(buf,nbr,master)) == NULL ) {
        switch( eno ) {
          case ENO_MALLOC:	/* low on mem */
            fprintf( ERROR_CHANNEL, "%s: Malloc() error.\n", pn );
            retval= RETVAL_ERROR; goto DIE_NOW;
            break;
          case ENO_MALEFMREP:	/* maleformed reply */
            fprintf( WARN_CHANNEL,
                     "%s: Ignoring maleformed response from %s:%d.\n",
                     pn, master->name, ntohs(master->sock.sin_port) );
            break;
          case ENO_OK:		/* 0 servers */
            break;
          default:
            fprintf( ERROR_CHANNEL, "%s: Bug detected! Email me at %s.\n",
                     pn, MY_EMAIL_ADDRESS );
            fprintf( ERROR_CHANNEL, "%s: create_hlist_from_mrep() eno %d.\n",
                     pn, eno );
            retval= RETVAL_BUG; goto DIE_NOW;
        }
      }
      nsg= count_hosts( alist );
      if( verbose >= 3 ) {
        fprintf( VERBOSE_CHANNEL, "%s: %3d servers (%s:%d)\n",
                 pn, nsg, master->name, ntohs(master->sock.sin_port) );
      }
      slist= extend_hlist( slist, alist );
    }
    if( verbose >= 2 ) {
      int noh;
      if( (noh=count_hosts_from_master(slist,NULL)) > 0 )
        fprintf( VERBOSE_CHANNEL, "%s: %4d game server%s from cmd-line.\n",
                 pn, noh, noh==1?"":"s" );
      for( master=mlist; master!=NULL; master=master->next ) {
        noh= count_hosts_from_master( slist, master );
        fprintf( VERBOSE_CHANNEL,
                 "%s: %4d protocol %02d game server%s from %s:%d.\n",
                 pn, noh, proto, noh==1?"":"s",
                 master->name, ntohs(master->sock.sin_port) );
      }
    }
    close( sock_fd );
  }

/* remove repeatedly occuring game servers from list */
  slist= correct_hlist( slist );
  if( verbose >= 2 ) {
    int n= count_hosts(slist);
    fprintf( VERBOSE_CHANNEL, "%s: %4d %sgame server%s.\n",
             pn, n, n>1?"different ":"", n==1?"":"s" );
  }

/* create missing game server names */
  slist= resolve_hlist( slist );

/* print uncontacted game server list (if requested) */
  if( list_hosts == 1 ) {
    if( verbose >= 1 ) {
      fprintf( VERBOSE_CHANNEL, "%s: Listing game servers.\n", pn );
    }
    fprint_hlist( STD_CHANNEL, "", slist );
    if( verbose >= 1 ) {
      fprintf( VERBOSE_CHANNEL, "%s: %d game servers listed.\n",
               pn, count_hosts(slist) );
    }
    goto DIE_NOW;
  }

/* what requests must be sent? */
  switch( list_hosts ) {
    case 1:	/* just list, don't contact */
      get_info= 0;
      get_status= 0;
      break;
    case 2:	/* list and contact */
      get_info= 1;
      get_status= 0;
      break;
    default:	/* print format dependent */
      if( format_requires_status(pformat) ) get_status= 1;
      else get_status= 0;
      if( format_requires_info(pformat) ) get_info= 1;
      else get_info= 0;
  }

/* get info and status from game server(s) (if requested) */
  if( get_info || get_status ) {
    struct host* server;
    int tnbs= 0;		/* total number of bytes sent */
    int tnbr= 0;		/* total number of bytes received */
    int tnirr= 0;		/* total n.o. info requests received */
    int tnsrr= 0;		/* total n.o. status requests received */
    int attempt;
    if( slist == NULL ) {
      fprintf( ERROR_CHANNEL, "%s: No valid game server known.\n", pn );
      retval= RETVAL_ERROR; goto DIE_NOW;
    }
    if( (sock_fd=create_socket(lport,&stimeout,sizeof(stimeout))) == -1 ) {
      retval= RETVAL_ERROR; goto DIE_NOW;
    }
    if( verbose >= 1  ) {
      int n= count_hosts(slist);
      fprintf( VERBOSE_CHANNEL,
               "%s: Requesting info/status from %d game server%s.\n",
               pn, n, n==1?"":"s" );
    }
    /* contact game server(s) */
    for( attempt=0; attempt<attempts; attempt++ ) {
      int nirts= 0;		/* number of info requests to send */
      int nsrts= 0;		/* number of status requests to send */
      int nirs= 0;		/* number of info requests sent */
      int nsrs= 0;		/* number of status requests sent */
      int nirr= 0;		/* number of info responses received */
      int nsrr= 0;		/* number of status responses received */
      if( verbose >= 2 ) {
        for( server=slist; server!=NULL; server=server->next ) {
          if( get_info && server->info == NULL ) nirts++;
          if( get_status && server->stat == NULL ) nsrts++;
        }
        if( nirts+nsrts)
          fprintf( VERBOSE_CHANNEL, "%s: Attempt %02d: Remaining %4d/%4d.\n",
                 pn, attempt+1, nirts, nsrts );
      }
      /* send info and status request to all servers */
      for( server=slist; server!=NULL; server=server->next ) {
        int nbs;
        if( get_info && server->info == NULL ) { /* send info request */
          char req[]= INFO_REQUEST;
          *(uint32_t*)req= MARKER;
          if( verbose >= 3 ) {
            fprintf( VERBOSE_CHANNEL, "%s: Requesting info from %s:%d.\n",
                     pn, server->name, ntohs(server->sock.sin_port) );
          }
          nbs= sendto(sock_fd,req,strlen(req),0, \
                       (struct sockaddr*)&server->sock,sizeof(server->sock) );
          if( nbs == -1 ) {
            fprintf( ERROR_CHANNEL, "%s: Sendto() failed for %s:%d. %s.\n",
                     pn, server->name, ntohs(server->sock.sin_port),
                     sys_errlist[errno] );
          }
          else {
            tnbs+= nbs;
            nirs++;
          }
        }
        if( get_status && server->stat == NULL ) { /* send status request */
          char req[]= STATUS_REQUEST;
          *(uint32_t*)req= MARKER;
          if( verbose >= 3 ) {
            fprintf( VERBOSE_CHANNEL, "%s: Requesting status from %s:%d.\n",
                     pn, server->name, ntohs(server->sock.sin_port) );
          }
          nbs= sendto(sock_fd,req,strlen(req),0, \
                       (struct sockaddr*)&server->sock,sizeof(server->sock) );
          if( nbs == -1 ) {
            fprintf( ERROR_CHANNEL, "%s: Sendto() failed for %s:%d. %s.\n",
                     pn, server->name, ntohs(server->sock.sin_port),
                     sys_errlist[errno] );
          }
          else {
            tnbs+= nbs;
            nsrs++;
          }
        }
      }
      /* listen to replies */
      while( nirs+nsrs ) {
        struct sockaddr_in ans_addr;
        socklen_t socklen= sizeof( ans_addr );
        int nbr;
        char s_resp[]= STATUS_RESPONSE;
        char i_resp[]= INFO_RESPONSE;
        *(uint32_t*)s_resp= MARKER;
        *(uint32_t*)i_resp= MARKER;
        nbr= recvfrom( sock_fd, buf, BUF_SIZE, 0, \
                       (struct sockaddr*)&ans_addr, &socklen );
        if( nbr == -1 ) {
          if( errno==EAGAIN ) {
            if( tnbr == 0 )
              fprintf( WARN_CHANNEL, "%s: No response from any server.\n", pn );
            break;
          }
          fprintf( ERROR_CHANNEL, "%s: Recvfrom() failed. %s.\n",
                   pn, sys_errlist[errno] );
          retval= RETVAL_ERROR; goto DIE_NOW;
        }
        for( server=slist; server!=NULL; server=server->next ) {
          if( memcmp(&server->sock,&ans_addr,socklen) == 0 ) break;
        }
        if( server == NULL ) {
          char ans_name[INET_ADDRSTRLEN]= "";
          int ans_port= ntohs( ans_addr.sin_port );
          inet_ntop( AF_INET, &ans_addr.sin_addr, ans_name, INET_ADDRSTRLEN );
          fprintf( WARN_CHANNEL, "%s: Ignoring dgram from %s:%d. %s.\n",
                   pn, ans_name, ans_port, "Game server unknown"  );
          continue;
        }
        if( nbr == BUF_SIZE ) {
          fprintf( ERROR_CHANNEL, "%s: Buffer too small. E-mail me at %s.\n",
                   pn, MY_EMAIL_ADDRESS );
          retval= RETVAL_BUG; goto DIE_NOW;
        }
        if( strncmp(buf,i_resp,strlen(i_resp)) == 0 ) {
          /* info response received */
          nirr++; tnirr++;
          if( verbose >= 3 ) {
            fprintf( VERBOSE_CHANNEL, "%s: Info response from %s:%d (%d).\n",
                     pn, server->name, ntohs(server->sock.sin_port), tnirr );
          }
          if( server->info != NULL ) {
            if( verbose >= 3 )
              fprintf( VERBOSE_CHANNEL, "%s: Server answered already!\n", pn );
            nirr--; tnirr--;
            continue;
          }
          if( (server->info=malloc(nbr+1-strlen(i_resp))) == NULL ) {
            fprintf( ERROR_CHANNEL, "%s: Malloc() error.\n", pn );
            retval= RETVAL_ERROR; goto DIE_NOW;
          }
          memcpy( server->info, buf+strlen(i_resp), nbr-strlen(i_resp) );
          server->info[nbr-strlen(i_resp)]= '\0';
        }
        else if( strncmp(buf,s_resp,strlen(s_resp)) == 0 ) {
          /* status response received */
          nsrr++; tnsrr++;
          if( verbose >= 3 ) {
            fprintf( VERBOSE_CHANNEL, "%s: Status response from %s:%d (%d).\n",
                     pn, server->name, ntohs(server->sock.sin_port), tnsrr );
          }
          if( server->stat != NULL ) {
            if( verbose >= 3 )
              fprintf( VERBOSE_CHANNEL, "%s: Server answered already!\n", pn );
            nsrr--; tnsrr--;
            continue;
          }
          if( (server->stat=malloc(nbr+1-strlen(s_resp))) == NULL ) {
            fprintf( ERROR_CHANNEL, "%s: Malloc() error.\n", pn );
            retval= RETVAL_ERROR; goto DIE_NOW;
          }
          memcpy( server->stat, buf+strlen(s_resp), nbr-strlen(s_resp) );
          server->stat[nbr-strlen(s_resp)]= '\0';
        }
        else {
          fprintf( WARN_CHANNEL, "%s: Ignoring maleformed dgram from %s:%d.\n",
                   pn, server->name, ntohs(server->sock.sin_port) );
        }
        tnbr+= nbr;
      }
      if( attempt==attempts-1 && nirr+nsrr>=5 ) {
        /* still so may answers... don't stop now */
        attempts++;
      }
    }
    if( verbose >= 1  ) {
      fprintf( VERBOSE_CHANNEL,
               "%s: Received info/status from %d/%d game servers.\n",
               pn, tnirr, tnsrr );
    }
    close( sock_fd );
  }

/* list all game servers which replied (if requested) */
  if( list_hosts == 2 ) {
    struct host* server;
    int nsl= 0;			/* number of servers listed */
    if( verbose >= 1 ) {
      fprintf( VERBOSE_CHANNEL, "%s: Listing game servers.\n", pn );
    }
    for( server=slist; server!=NULL; server=server->next ) {
      if( server->info != NULL || server->stat != NULL ) {
        fprintf( STD_CHANNEL, "%s:%d\n",
                 server->name, ntohs(server->sock.sin_port) );
        nsl++;
      }
    }
    if( verbose >= 1 ) {
      fprintf( VERBOSE_CHANNEL, "%s: %d game servers listed.\n",
               pn, nsl );
    }
    goto DIE_NOW;
  }

/* print results for all servers */
  {
    struct host* server;
    for( server=slist; server!=NULL; server=server->next ) {
      fprintf_response( STD_CHANNEL, pformat, color_mode, server );
/*
fprintf( STD_CHANNEL, "DEBUG: %s\n", server->info );
fprintf( STD_CHANNEL, "DEBUG: %s\n", server->stat );
*/
    }
  }


DIE_NOW:
  free_hlist( mlist );
  free_hlist( slist );
  close( sock_fd );
  exit( retval );
}
