#undef	DEBUG

#include <stdio.h>
#include <errno.h>
#include <unistd.h>		/* getopt(), read(), write(), close(), STDIN_FILENO */
#include <stdlib.h>		/* strtol(), malloc() */
#include <ctype.h>		/* tolower() */
#include <sys/types.h>		/* open(), socket(), connect(), sendto() */
#include <sys/stat.h>		/* open() */
#include <sys/socket.h>		/* socket(), connect() */
#include <fcntl.h>		/* open(), O_RDONLY */
#include <string.h>		/* strlen(), memset() */
#include <netdb.h>		/* gethostbyname() */
#include <netinet/in.h>		/* htons(), struct sockaddr_in, struct in_addr */
#include <arpa/inet.h>		/* inet_addr() */
#include <time.h>		/* nanosleep() */


#define VERBOSE_CHANNEL	stderr		/* additional info */
#define	WARN_CHANNEL	stderr		/* warnings */
#define	STD_CHANNEL	stdout		/* normal output */
#define HELP_CHANNEL	stdout
#define	VERSION_CHANNEL	stdout
#define ERROR_CHANNEL	stderr		/* errors */
#define DEBUG_CHANNEL	stderr

#define	DEFAULT_MNAME		"wolfmaster.idsoftware.com"
#define	DEFAULT_MPORT		27950
#define	DEFAULT_SNAME		"absinth.dyndns.org"
#define	DEFAULT_SPORT		27960
#define	DEFAULT_LPORT		27960
#define	DEFAULT_PROTO		58
#define	DEFAULT_MTIMEOUT	10
#define	DEFAULT_STIMEOUT	2
#define	DEFAULT_ATTEMPTS	5
#define	DEFAULT_PFORMAT		"%H:%P\\t%n\\t%m\\t%#/%^\\n"
#define	DEFAULT_COLOR_MODE	2

/*
 * special wolfinfo print format characters
 */
/* general */
#define	GFORMAT_HOST		'H'
#define	GFORMAT_PORT		'P'
#define	GFORMAT_LINE		'L'
/* info */
#define	IFORMAT_ALL		'I'
#define	IFORMAT_PROTO		'o'
#define	IFORMAT_SERVER		'n'
#define	IFORMAT_MAP		'm'
#define	IFORMAT_CLIENTS		'#'
#define	IFORMAT_MAXCLIENTS	'^'
#define	IFORMAT_GAMETYPE	'g'
#define	IFORMAT_PURE
#define	IFORMAT_MAXPING
#define	IFORMAT_ANON
#define	IFORMAT_GAMESKILL	's'
#define	IFORMAT_FRIENDLY	'f'
#define	IFORMAT_MAXLIVES	'M'
#define	IFORMAT_TOURNEY		'u'
#define	IFORMAT_PUNK		'b'
/* status */
#define	SFORMAT_ALL		'S'
#define	SFORMAT_PROTO
#define	SFORMAT_SERVER
#define	SFORMAT_MAP
#define	SFORMAT_MAXCLIENTS
#define	SFORMAT_GAMETYPE
#define	SFORMAT_MINPING
#define	SFORMAT_MAXPING
#define	SFORMAT_ANON
#define	SFORMAT_GAMESKILL
#define	SFORMAT_FRIENDLY
#define	SFORMAT_MAXLIVES
#define	SFORMAT_VERSION		'v'
#define	SFORMAT_SOL_CT
#define	SFORMAT_LIE_CT
#define	SFORMAT_ENG_CT
#define	SFORMAT_MED_CT
#define	SFORMAT_BLUE_LT
#define	SFORMAT_RED_LT
#define	SFORMAT_GAMENAME
#define	SFORMAT_MINGCLIENTS
#define	SFORMAT_MAXGCLIENTS
#define	SFORMAT_FLOOD
#define	SFORMAT_MAXRATE
#define	SFORMAT_BREATH
#define	SFORMAT_PRIVCLIENTS
#define	SFORMAT_TIMELIMIT	't'
#define	SFORMAT_PLAYERS		'p'

/* info response variables (lowercase letters!) */
#define	IRESP_PROTO		"\\protocol\\"
#define	IRESP_SERVER		"\\hostname\\"
#define	IRESP_MAP		"\\mapname\\"
#define	IRESP_CLIENTS		"\\clients\\"
#define	IRESP_MAXCLIENTS	"\\sv_maxclients\\"
#define	IRESP_GAMETYPE		"\\gametype\\"
#define	IRESP_PURE		"\\pure\\"
#define	IRESP_MAXPING		"\\maxping\\"
#define	IRESP_ANON		"\\sv_allowanonymous\\"
#define	IRESP_GAMESKILL		"\\gameskill\\"
#define	IRESP_FRIENDLY		"\\friendlyfire\\"
#define	IRESP_MAXLIVES		"\\maxlives\\"
#define	IRESP_TOURNEY		"\\tourney\\"
#define	IRESP_PUNK		"\\punkbuster\\"
/* status response variables (lowercase letters!) */
#define	SRESP_PROTO		"\\protocol\\"
#define	SRESP_SERVER		"\\sv_hostname\\"
#define	SRESP_MAP		"\\mapname\\"
#define	SRESP_MAXCLIENTS	"\\g_maxclients\\"
#define	SRESP_GAMETYPE		"\\g_gametype\\"
#define	SRESP_MINPING		"\\sv_minping\\"
#define	SRESP_MAXPING		"\\sv_maxping\\"
#define	SRESP_ANON		"\\sv_allowanonymous\\"
#define	SRESP_GAMESKILL		"\\g_gameskill\\"
#define	SRESP_FRIENDLY		"\\g_friendlyfire\\"
#define	SRESP_MAXLIVES		"\\g_maxlives\\"
#define	SRESP_VERSION		"\\version\\"
#define	SRESP_SOL_CT		"\\g_soldierChargeTime\\"
#define	SRESP_LIE_CT		"\\g_LTChargeTime\\"
#define	SRESP_ENG_CT		"\\g_engineerChargeTime\\"
#define	SRESP_MED_CT		"\\g_medicChargeTime\\"
#define	SRESP_BLUE_LT		"\\g_bluelimbotime\\"
#define	SRESP_RED_LT		"\\g_redlimbotime\\"
#define	SRESP_GAMENAME		"\\gamename\\"
#define	SRESP_MINGCLIENTS	"\\g_mingameclients\\"
#define	SRESP_MAXGCLIENTS	"\\g_maxgameclients\\"
#define	SRESP_FLOOD		"\\sv_floodprotect\\"
#define	SRESP_MAXRATE		"\\sv_maxrate\\"
#define	SRESP_BREATH		"\\g_enablebreath\\"
#define	SRESP_PRIVCLIENTS	"\\sv_privateClients\\"
#define	SRESP_TIMELIMIT		"\\timelimit\\"


#define	INFO_REQUEST		"XXXXgetinfo xxx"
#define	INFO_RESPONSE		"XXXXinfoResponse\n"
#define	STATUS_REQUEST		"XXXXgetstatus"
#define	STATUS_RESPONSE		"XXXXstatusResponse\n"

#define	MARKER		0xffffffff
#define	ESC_CHAR	'\\'
#define	END_STR		"EOT"

#define	COLOR_ESC_CHAR	'^'

#define	RETVAL_OK		0
#define	RETVAL_BUG		1
#define	RETVAL_ERROR		2

#define	ENO_OK		0
#define	ENO_MALLOC	1
#define	ENO_INVCALL	2	/* invalid call */
#define	ENO_INVPORT	3	/* invalid port */
#define	ENO_MALEFMREP	4	/* maleformed master reply */

#define	BUF_SIZE	0x1000

#define VERSION_NUMBER		"0.02.01"
#define DATE_OF_LAST_MOD	"2002-03-21"
#define	MY_EMAIL_ADDRESS	"Steffen Solyga <solyga@absinth.net>"

#define	N_( A )			A
/* taken from resolv/herror.c of glibc-2.0.7 */
const char *h_errlist[] = {
        N_("Resolver Error 0 (no error)"),
        N_("Unknown host"),                     /* 1 HOST_NOT_FOUND */
        N_("Host name lookup failure"),         /* 2 TRY_AGAIN */
        N_("Unknown server error"),             /* 3 NO_RECOVERY */
        N_("No address associated with name"),  /* 4 NO_ADDRESS */
/* my own extension (yes, quite terrible) */
	N_("No IPv4 address available"),	/* 5 my extension for IPv6 */
};

struct host {
  struct host* next;		/* next host in list */
  struct host* where;		/* if game server: master */
  char* name;
  struct sockaddr_in sock;
  char* info;			/* game server info (raw) */
  char* stat;			/* game server status (raw) */
};


char* color_table[]= {
	"02;30",	/* black */
	"02;31",	/* red */
	"02;32",	/* green */
	"02;33",	/* yellow */
	"02;34",	/* blue */
	"02;36",	/* light blue */
	"02;35",	/* pink */
	"02;37",	/* white */
	"02;30",	/* black (bold) */
	"01;31",	/* red (bold) */
	"01;32",	/* green (bold) */
	"01;33",	/* yellow (bold) */
	"01;34",	/* blue (bold) */
	"01;36",	/* light blue (bold) */
	"01;35",	/* pink (bold) */
	"01;37",	/* white (bold) */
	"00"		/* normal */
};
